@@ -1,17 +1,18 @@
-#![feature(const_fn_trait_bound)]
-#![feature(const_mut_refs)]
-#![feature(const_fn_fn_ptr_basics)]
-#![feature(asm)]
-#![feature(decl_macro)]
-#![feature(generic_const_exprs)]
-#![feature(const_maybe_uninit_as_ptr)]
 #![feature(const_ptr_offset_from)]
-#![feature(const_raw_ptr_deref)]
+#![feature(generic_const_exprs)]
 #![feature(const_refs_to_cell)]
-#![feature(slice_ptr_len)]
+#![feature(const_trait_impl)]
 #![feature(naked_functions)]
+#![feature(const_mut_refs)]
+#![feature(slice_ptr_len)]
+#![feature(decl_macro)]
+#![feature(asm_const)]
+#![feature(asm_sym)]
 #![deny(unsafe_op_in_unsafe_fn)]
-#![deny(unsupported_naked_functions)]
+#![cfg_attr(
+    feature = "doc",
+    doc(html_logo_url = "https://r3-os.github.io/r3/logo-small.svg")
+)]
 #![doc = include_str!("./lib.md")]
 #![no_std]
 
@@ -19,7 +20,7 @@
 #[doc = include_str!("../CHANGELOG.md")]
 pub mod _changelog_ {}
 
-/// The [`r3::kernel::PortThreading`] implementation.
+/// The [`r3_kernel::PortThreading`] implementation.
 #[doc(hidden)]
 pub mod threading {
     pub mod cfg;
@@ -35,7 +36,7 @@ pub mod rt {
     pub mod imp;
 }
 
-/// The tickful [`r3::kernel::PortTimer`] implementation based on SysTick.
+/// The tickful [`r3_kernel::PortTimer`] implementation based on SysTick.
 #[doc(hidden)]
 pub mod systick_tickful {
     pub mod cfg;
@@ -51,7 +52,13 @@ pub use self::{systick_tickful::cfg::*, threading::cfg::*};
 pub extern crate core;
 /// Used by `use_port!`
 #[doc(hidden)]
-pub extern crate r3;
+pub extern crate r3_core;
+/// Used by `use_port!`
+#[doc(hidden)]
+pub extern crate r3_kernel;
+/// Used by `use_port!`
+#[doc(hidden)]
+pub extern crate r3_portkit;
 /// Used by `use_port!`
 #[doc(hidden)]
 #[cfg(target_os = "none")]
