@@ -1,37 +1,34 @@
-use core::{cell::UnsafeCell, mem::MaybeUninit, slice};
+use core::{
+    cell::UnsafeCell,
+    mem::MaybeUninit,
+    slice,
+    sync::atomic::{compiler_fence, Ordering},
+};
 use memoffset::offset_of;
-use r3::{
+use r3_core::{
     kernel::{
-        ClearInterruptLineError, EnableInterruptLineError, InterruptNum, InterruptPriority,
-        PendInterruptLineError, Port, PortToKernel, QueryInterruptLineError,
-        SetInterruptLinePriorityError, TaskCb,
+        traits, ClearInterruptLineError, EnableInterruptLineError, InterruptNum, InterruptPriority,
+        PendInterruptLineError, QueryInterruptLineError, SetInterruptLinePriorityError,
     },
-    prelude::*,
-    utils::{Init, ZeroInit},
-};
-use r3_portkit::{
-    pptext::pp_asm,
-    sym::{sym_static, SymStaticExt},
+    utils::Init,
 };
+use r3_kernel::{KernelTraits, Port, PortToKernel, System, TaskCb};
+use r3_portkit::{pptext::pp_asm, sym::sym_static};
 
 use crate::{
     ThreadingOptions, INTERRUPT_EXTERNAL0, INTERRUPT_NUM_RANGE, INTERRUPT_PRIORITY_RANGE,
     INTERRUPT_SYSTICK,
 };
 
-/// Implemented on a system type by [`use_port!`].
+/// Implemented on a kernel trait type by [`use_port!`].
 ///
 /// # Safety
 ///
 /// Only meant to be implemented by [`use_port!`].
 pub unsafe trait PortInstance:
-    Kernel + Port<PortTaskState = TaskState> + ThreadingOptions
+    KernelTraits + Port<PortTaskState = TaskState> + ThreadingOptions
 {
-    sym_static!(static PORT_STATE: SymStatic<State> = zeroed!());
-
-    fn port_state() -> &'static State {
-        sym_static(Self::PORT_STATE).as_ref()
-    }
+    sym_static!(#[sym(p_port_state)] fn port_state() -> &State);
 }
 /// Converts [`InterruptNum`] to [`cortex_m::interrupt::Nr`].
 #[derive(Clone, Copy)]
@@ -45,9 +42,9 @@ unsafe impl cortex_m::interrupt::Nr for Int {
 }
 
 pub struct State {
-    /// Stores the value of `System::state().running_task_ptr()` so that it can
+    /// Stores the value of `Traits::state().running_task_ptr()` so that it can
     /// be accessed in naked functions. This field is actually of type
-    /// `*mut Option<&'static TaskCb<System>>`.
+    /// `*mut Option<&'static TaskCb<Traits>>`.
     running_task_ptr: UnsafeCell<*mut ()>,
 }
 
@@ -56,7 +53,12 @@ impl State {
 }
 
 unsafe impl Sync for State {}
-unsafe impl ZeroInit for State {}
+
+impl Init for State {
+    const INIT: Self = Self {
+        running_task_ptr: UnsafeCell::new(core::ptr::null_mut()),
+    };
+}
 
 #[derive(Debug)]
 #[repr(C)]
@@ -67,16 +69,16 @@ pub struct TaskState {
 unsafe impl Sync for TaskState {}
 
 impl Init for TaskState {
+    #[allow(clippy::declare_interior_mutable_const)] // it's intentional
     const INIT: Self = Self {
         sp: UnsafeCell::new(0),
     };
 }
 
 impl State {
-    pub unsafe fn port_boot<System: PortInstance>(&self) -> ! {
-        unsafe { self.enter_cpu_lock::<System>() };
-
-        unsafe { *self.running_task_ptr.get() = System::state().running_task_ptr() as *mut () };
+    #[inline(always)]
+    pub unsafe fn port_boot<Traits: PortInstance>(&self) -> ! {
+        unsafe { self.enter_cpu_lock::<Traits>() };
 
         // Claim the ownership of `Peripherals`
         let mut peripherals = unsafe { cortex_m::Peripherals::steal() };
@@ -94,11 +96,15 @@ impl State {
 
         // Safety: We are a port, so it's okay to call this
         unsafe {
-            <System as PortToKernel>::boot();
+            <Traits as PortToKernel>::boot();
         }
     }
 
-    pub unsafe fn dispatch_first_task<System: PortInstance>(&'static self) -> ! {
+    #[inline(always)]
+    pub unsafe fn dispatch_first_task<Traits: PortInstance>(&'static self) -> ! {
+        // [tag:running_task_ptr_set_in_dft]
+        unsafe { *self.running_task_ptr.get() = Traits::state().running_task_ptr() as *mut () };
+
         // Pend PendSV
         cortex_m::peripheral::SCB::set_pendsv();
 
@@ -108,7 +114,7 @@ impl State {
         //
         // Safety: `CONTROL.SPSEL == 0`, Thread mode (entailed by the boot
         // context), CPU Lock active
-        unsafe { Self::idle_task::<System>() };
+        unsafe { Self::idle_task::<Traits>() };
     }
 
     /// Reset MSP to `interrupt_stack_top()`, release CPU Lock, and start
@@ -118,10 +124,10 @@ impl State {
     ///
     /// `CONTROL.SPSEL == 0`, Thread mode, CPU Lock active
     #[inline(never)]
-    unsafe extern "C" fn idle_task<System: PortInstance>() -> ! {
+    unsafe extern "C" fn idle_task<Traits: PortInstance>() -> ! {
         // Find the top of the interrupt stack
         // Safety: Only the port can call this method
-        let msp_top = unsafe { System::interrupt_stack_top() };
+        let msp_top = unsafe { Traits::interrupt_stack_top() };
 
         pp_asm!("
             # Reset MSP to the top of the stack, effectively discarding the
@@ -144,7 +150,7 @@ impl State {
             msp_top = in(reg) msp_top,
         );
 
-        if System::USE_WFI {
+        if Traits::USE_WFI {
             pp_asm!(
                 "
             0:
@@ -164,14 +170,25 @@ impl State {
         }
     }
 
-    pub unsafe fn yield_cpu<System: PortInstance>(&'static self) {
+    pub unsafe fn yield_cpu<Traits: PortInstance>(&'static self) {
+        // Ensure preceding memory operations are visible to the PendSV handler
+        compiler_fence(Ordering::Release);
+
         // Safety: See `use_port!`
         cortex_m::peripheral::SCB::set_pendsv();
+
+        // Technically this DSB isn't required for correctness, but ensures
+        // PendSV is taken before the next operation.
+        cortex_m::asm::dsb();
+
+        // Ensure the PendSV handler's memory operations are visible to us
+        compiler_fence(Ordering::Acquire);
     }
 
-    pub unsafe fn exit_and_dispatch<System: PortInstance>(
+    #[inline(always)]
+    pub unsafe fn exit_and_dispatch<Traits: PortInstance>(
         &'static self,
-        _task: &'static TaskCb<System>,
+        _task: &'static TaskCb<Traits>,
     ) -> ! {
         // Pend PendSV
         cortex_m::peripheral::SCB::set_pendsv();
@@ -200,7 +217,7 @@ impl State {
         "   } else {                                                        "
                 b {idle_task}
         "   },
-            idle_task = sym Self::idle_task::<System>,
+            idle_task = sym Self::idle_task::<Traits>,
             options(noreturn),
         );
     }
@@ -213,7 +230,7 @@ impl State {
     ///    registers must contain the values from the background context.
     ///
     #[naked]
-    pub unsafe extern "C" fn handle_pend_sv<System: PortInstance>() {
+    pub unsafe extern "C" fn handle_pend_sv<Traits: PortInstance>() {
         // Precondition:
         //  - `EXC_RETURN.Mode == 1` - Exception was taken in Thread mode. This
         //    is true because PendSV is configured with the lowest priority.
@@ -224,14 +241,14 @@ impl State {
         //    context is the idle task, the exception frame should have been
         //    stacked to MSP.
 
-        extern "C" fn choose_next_task<System: PortInstance>() {
+        extern "C" fn choose_next_task<Traits: PortInstance>() {
             // Choose the next task to run
-            unsafe { State::enter_cpu_lock_inner::<System>() };
+            unsafe { State::enter_cpu_lock_inner::<Traits>() };
 
             // Safety: CPU Lock active
-            unsafe { System::choose_running_task() };
+            unsafe { Traits::choose_running_task() };
 
-            unsafe { State::leave_cpu_lock_inner::<System>() };
+            unsafe { State::leave_cpu_lock_inner::<Traits>() };
         }
 
         pp_asm!("
@@ -256,7 +273,8 @@ impl State {
             #
             #    <r0 = &running_task>
 
-            ldr r0, ={PORT_STATE}_
+            ldr r0, ={p_port_state}_
+            ldr r0, [r0]
             ldr r0, [r0, #{OFFSET_RUNNING_TASK_PTR}]
 
             ldr r1, [r0]                                                    "
@@ -374,24 +392,47 @@ impl State {
             msr control, r0
             bx lr
         ",
-            choose_next_task = sym choose_next_task::<System>,
-            PORT_STATE = sym System::PORT_STATE,
+            choose_next_task = sym choose_next_task::<Traits>,
+            p_port_state = sym Traits::p_port_state,
             OFFSET_RUNNING_TASK_PTR = const Self::OFFSET_RUNNING_TASK_PTR,
             options(noreturn),
         );
     }
 
     #[inline(always)]
-    pub unsafe fn enter_cpu_lock<System: PortInstance>(&self) {
-        unsafe { Self::enter_cpu_lock_inner::<System>() };
+    pub unsafe fn enter_cpu_lock<Traits: PortInstance>(&self) {
+        unsafe { Self::enter_cpu_lock_inner::<Traits>() };
     }
 
     #[inline(always)]
-    unsafe fn enter_cpu_lock_inner<System: PortInstance>() {
+    unsafe fn enter_cpu_lock_inner<Traits: PortInstance>() {
         #[cfg(not(any(armv6m, armv8m_base)))]
-        if System::CPU_LOCK_PRIORITY_MASK > 0 {
+        if Traits::CPU_LOCK_PRIORITY_MASK > 0 {
             // Set `BASEPRI` to `CPU_LOCK_PRIORITY_MASK`
-            unsafe { cortex_m::register::basepri::write(System::CPU_LOCK_PRIORITY_MASK) };
+            unsafe { cortex_m::register::basepri::write(Traits::CPU_LOCK_PRIORITY_MASK) };
+
+            // Synchronize with the previous owner of CPU Lock.
+            //
+            // The semantics of `compiler_fence` we need here for this to work
+            // might be more strict than what `compiler_fence`'s documentation
+            // ("the compiler may be disallowed from moving reads or writes from
+            // before or after the call to the other side of the call to
+            // `compiler_fence`"; but it doesn't say it can't be moved past
+            // `basepri::write`, which has `options(nomem)`[2] and therefore is
+            // not a memory operation) and the C++ memory model ("memory
+            // synchronization ordering of non-atomic and relaxed atomic
+            // accesses"[3] but there's no atomic access here) say. But on the
+            // other hand, there's a code comment[1] from the `cortex-m` package
+            // (maintained by Rust's official embedded devices WG[4]) suggesting
+            // that `compiler_fence` can in fact prevent the reordering as
+            // intended. We're going to take their word for it since we are
+            // using this package anyway.
+            //
+            // [1]: https://github.com/rust-embedded/cortex-m/blob/92552c73d3b56dc86007450633950d16ebe0e495/asm/inline.rs#L36
+            // [2]: https://github.com/rust-embedded/cortex-m/blob/92552c73d3b56dc86007450633950d16ebe0e495/asm/inline.rs#L243
+            // [3]: https://en.cppreference.com/w/cpp/atomic/atomic_signal_fence
+            // [4]: https://github.com/rust-embedded/wg
+            compiler_fence(Ordering::Acquire);
             return;
         }
 
@@ -400,14 +441,17 @@ impl State {
     }
 
     #[inline(always)]
-    pub unsafe fn leave_cpu_lock<System: PortInstance>(&'static self) {
-        unsafe { Self::leave_cpu_lock_inner::<System>() };
+    pub unsafe fn leave_cpu_lock<Traits: PortInstance>(&'static self) {
+        unsafe { Self::leave_cpu_lock_inner::<Traits>() };
     }
 
     #[inline(always)]
-    unsafe fn leave_cpu_lock_inner<System: PortInstance>() {
+    unsafe fn leave_cpu_lock_inner<Traits: PortInstance>() {
         #[cfg(not(any(armv6m, armv8m_base)))]
-        if System::CPU_LOCK_PRIORITY_MASK > 0 {
+        if Traits::CPU_LOCK_PRIORITY_MASK > 0 {
+            // Synchronize with the next owner of CPU Lock.
+            compiler_fence(Ordering::Release);
+
             // Set `BASEPRI` to `0` (no masking)
             unsafe { cortex_m::register::basepri::write(0) };
             return;
@@ -417,9 +461,9 @@ impl State {
         unsafe { cortex_m::interrupt::enable() };
     }
 
-    pub unsafe fn initialize_task_state<System: PortInstance>(
+    pub unsafe fn initialize_task_state<Traits: PortInstance>(
         &self,
-        task: &'static TaskCb<System>,
+        task: &'static TaskCb<Traits>,
     ) {
         let stack = task.attr.stack.as_ptr();
         let mut sp = (stack as *mut u8).wrapping_add(stack.len()) as *mut MaybeUninit<u32>;
@@ -435,7 +479,7 @@ impl State {
         };
 
         // R0: Parameter to the entry point
-        exc_frame[0] = MaybeUninit::new(task.attr.entry_param as u32);
+        exc_frame[0] = unsafe { core::mem::transmute(task.attr.entry_param) };
         // R1-R3, R12: Uninitialized
         if preload_all {
             exc_frame[1] = MaybeUninit::new(0x01010101);
@@ -444,7 +488,8 @@ impl State {
             exc_frame[4] = MaybeUninit::new(0x12121212);
         }
         // LR: The return address
-        exc_frame[5] = MaybeUninit::new(System::exit_task as usize as u32);
+        exc_frame[5] =
+            MaybeUninit::new(<System<Traits> as traits::KernelBase>::raw_exit_task as usize as u32);
         // PC: The entry point - The given function pointer has its LSB set to
         // signify that the target is a Thumb function (that's the only valid
         // mode on Arm-M) as required by the BLX instruction. In an exception
@@ -491,22 +536,43 @@ impl State {
     }
 
     #[inline(always)]
-    pub fn is_cpu_lock_active<System: PortInstance>(&self) -> bool {
+    pub fn is_cpu_lock_active<Traits: PortInstance>(&self) -> bool {
         #[cfg(not(any(armv6m, armv8m_base)))]
-        if System::CPU_LOCK_PRIORITY_MASK > 0 {
+        if Traits::CPU_LOCK_PRIORITY_MASK > 0 {
             return cortex_m::register::basepri::read() != 0;
         }
 
         cortex_m::register::primask::read().is_inactive()
     }
 
-    pub fn is_task_context<System: PortInstance>(&self) -> bool {
+    pub fn is_task_context<Traits: PortInstance>(&self) -> bool {
         // All tasks use PSP. The idle task is the exception, but user
         // code cannot run in the idle task, so we can ignore this.
         cortex_m::register::control::read().spsel() == cortex_m::register::control::Spsel::Psp
     }
 
-    pub fn set_interrupt_line_priority<System: PortInstance>(
+    #[inline]
+    pub fn is_interrupt_context<Traits: PortInstance>(&self) -> bool {
+        // `IPSR.Exception != 0`
+        unsafe {
+            let ipsr: u32;
+            pp_asm!(
+                "mrs {}, ipsr",
+                out(reg) ipsr,
+                options(nomem, preserves_flags, nostack),
+            );
+            (ipsr & ((1u32 << 9) - 1)) != 0
+        }
+    }
+
+    #[inline]
+    pub fn is_scheduler_active<Traits: PortInstance>(&self) -> bool {
+        // `runnin_task_ptr` is assigned by `dispatch_first_task`
+        // [ref:running_task_ptr_set_in_dft]
+        unsafe { !(*self.running_task_ptr.get()).is_null() }
+    }
+
+    pub fn set_interrupt_line_priority<Traits: PortInstance>(
         &'static self,
         num: InterruptNum,
         priority: InterruptPriority,
@@ -535,7 +601,7 @@ impl State {
     }
 
     #[inline]
-    pub fn enable_interrupt_line<System: PortInstance>(
+    pub fn enable_interrupt_line<Traits: PortInstance>(
         &'static self,
         num: InterruptNum,
     ) -> Result<(), EnableInterruptLineError> {
@@ -551,7 +617,7 @@ impl State {
     }
 
     #[inline]
-    pub fn disable_interrupt_line<System: PortInstance>(
+    pub fn disable_interrupt_line<Traits: PortInstance>(
         &self,
         num: InterruptNum,
     ) -> Result<(), EnableInterruptLineError> {
@@ -566,7 +632,7 @@ impl State {
     }
 
     #[inline]
-    pub fn pend_interrupt_line<System: PortInstance>(
+    pub fn pend_interrupt_line<Traits: PortInstance>(
         &'static self,
         num: InterruptNum,
     ) -> Result<(), PendInterruptLineError> {
@@ -584,7 +650,7 @@ impl State {
     }
 
     #[inline]
-    pub fn clear_interrupt_line<System: PortInstance>(
+    pub fn clear_interrupt_line<Traits: PortInstance>(
         &self,
         num: InterruptNum,
     ) -> Result<(), ClearInterruptLineError> {
@@ -602,7 +668,7 @@ impl State {
     }
 
     #[inline]
-    pub fn is_interrupt_line_pending<System: PortInstance>(
+    pub fn is_interrupt_line_pending<Traits: PortInstance>(
         &self,
         num: InterruptNum,
     ) -> Result<bool, QueryInterruptLineError> {
@@ -619,10 +685,10 @@ impl State {
 }
 
 /// Used by `use_port!`
-pub const fn validate<System: PortInstance>() {
+pub const fn validate<Traits: PortInstance>() {
     #[cfg(any(armv6m, armv8m_base))]
     assert!(
-        System::CPU_LOCK_PRIORITY_MASK == 0,
+        Traits::CPU_LOCK_PRIORITY_MASK == 0,
         "`CPU_LOCK_PRIORITY_MASK` must be zero because the target architecture \
          does not have a BASEPRI register"
     );
